/**
 * MPQAudioFragment.java: Fragment to tie together audio related components, 
 * 							presenting UI widgets, including a waveform View, 
 * 							and starting a service to record and play audio using worker threads.
 *            Copyright (C) 2012,2013 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-12-17 donfede@casagrau.org
 *
 * $Id: MPQAudioFragment.java,v 1.13 2013-02-09 20:14:01 donfede Exp $
 */

package org.casagrau.mpq.audio_service;

import org.casagrau.mpq.mpq_android.R;
import org.casagrau.mpq.mpq_android.SessionsDBProvider.SessionsDB;

import android.app.Activity;
import android.app.ProgressDialog;
import android.content.ComponentName;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.ServiceConnection;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.os.IBinder;
import android.support.v4.app.Fragment;
import android.util.DisplayMetrics;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.ViewGroup;
import android.widget.Button;
import android.widget.SeekBar;
import android.widget.TextView;

import com.ringdroid.soundfile.CheapSoundFile;

/**
 * Fragment to tie together audio related components.  Presents the UI widgets, 
 * including a waveform View (ported from the Ringdroid project), and uses a 
 * foreground running android service, to record and play audio.  Worker threads 
 * do the actual audio processing, for an optimal experience that does not slow 
 * down the UI thread.
 */
public class MPQAudioFragment extends Fragment {
	private static final String LOGTAG = "MPQAudioFragment";
	
	private Activity mAct;				// Caching variable with the activity, for context and related methods.
    private Handler mFragHandler;		// Handler for the Fragment.  // FIXME: not clear if this is shared with the Activity or not... (same thread, but ...?) -- alternately, do we grab the handler from the activity???

	// Audio UI components
	private Button mPracticeButton; 	// buttonPracticeToggle
	private Button mTCButton; 			// buttonAUI_TC -- transport control: record, play, pause toggle
	private Button mStopButton;			// buttonAUI_STOP
	private SeekBar mVolControl;		// widget to set the volume (either speaker or mic, depending on the state)
	private TextView mLabelAudioPos;	// label for the cursor position
	private TextView mLabelAudioLen;	// label_total_counter , representing the total file length (in hh:mm:ss)
	private TextView mLabelVolControl; 	// label for the volume control (ex: mic or volume)
	private TextView mLabelVolValue;	// label to display the current volume value
	private TextView mLabelFilename;	// label to display the current filename
	
	// Member attributes used update UI widgets, keeping track of state and values of the audio service. 
	private MPQAudioService mAudioService;		// Handle to the audio service, received from a binding connection.
	private boolean mAudioServiceBound = false;	// Flag to track if the audio service is bound.
	private int mAudioServiceState = MPQAudioService.FLG_SRV_STATE_INIT;	// Audio service state.
	private String mAudioServiceName;										// String with filename of the current audio managed by the audio service.
	private long mAudioServicePos;											// Audio service cursor position, in milliseconds.
	private long mAudioServiceLen;										// Audio service *Audio* length, in milliseconds.
	private int mAudioServiceVol;						
		
	// Member attributes related to WaveformView
	private WaveformView mAudioWaveformView;// Handle to the waveform widget.
    private int mAudioWaveformWidth;		// Track the width of the waveform. // FIXME: currently, this is not used - review, possibly delete...
    private boolean mAudioWaveformKeyDown;  // FIXME: review if we want to keep this...
    private int mAudioWaveformOffset;
    private int mAudioWaveformOffsetGoal;
    private int mAudioWaveformStartPos;
    private int mAudioWaveformEndPos;
    private int mAudioWaveformMaxPos; // FIXME: should probably be long... and document this (once it works...)
    private float mAudioWaveformDensity;
    private CheapSoundFile mAudioWaveformSoundFile;	// Handle to "CheapSoundFile", used to by waveform widget.
//////    private long mAudioWaveformLoadingStartTime;
    private long mAudioWaveformLoadingLastUpdateTime;
    private boolean mAudioWaveformLoadingKeepGoing;
    private ProgressDialog mAudioWaveformProgressDialog;
	
	
	
	public MPQAudioFragment() {
		super();
        Log.d(LOGTAG, "CONSTRUCTOR() 00");
	}
	
	@Override
	public void onAttach(Activity activity) {
		super.onAttach(activity);
		Log.d(LOGTAG, "onAttach() 00 getting activity");
		mAct = activity;
	}
	
    @Override
    public View onCreateView(LayoutInflater inflater, ViewGroup container,
                             Bundle savedInstanceState) {
    	super.onCreateView(inflater, container, savedInstanceState);
		Log.d(LOGTAG, "onCreateView() 00");
        mFragHandler = new Handler();
        
        // Inflate the layout for this fragment
        return inflater.inflate(R.layout.fragment_audio, container, false);
    }
	
    @Override
    public void onStart() {
    	super.onStart();
    	Log.d(LOGTAG, "onStart() 00");
    	Intent intent = new Intent(mAct, MPQAudioService.class);
		intent.setAction(MPQAudioService.ACTION_BIND);
		mAct.bindService(intent, mAServiceConnection, Context.BIND_AUTO_CREATE);
    	Log.d(LOGTAG, "onStart() 99");
    }
	
    @Override
    public void onResume() {
    	super.onResume();
    	Log.d(LOGTAG, "onResume() 00");		
    	
		// Bind Audio UI widget listeners.
		mPracticeButton = (Button) mAct.findViewById(R.id.buttonPracticeToggle);
		mPracticeButton.setOnClickListener(clickerPractice);
		mTCButton = (Button) mAct.findViewById(R.id.buttonAUI_TC);
		mTCButton.setOnClickListener(clickerTC);		
		mStopButton = (Button) mAct.findViewById(R.id.buttonAUI_STOP);
		mStopButton.setOnClickListener(clickerStop);
		mVolControl = (SeekBar) mAct.findViewById(R.id.seek_vol_control);
		mVolControl.setVisibility(View.GONE);
//		mVolControl.setOnSeekBarChangeListener(seekerVolControl);
		mLabelAudioPos = (TextView) mAct.findViewById(R.id.label_cursor_counter);
		mLabelAudioLen = (TextView) mAct.findViewById(R.id.label_total_counter);
		mLabelVolControl = (TextView) mAct.findViewById(R.id.label_vol_control);
		mLabelVolValue = (TextView) mAct.findViewById(R.id.label_vol_value);
		mLabelFilename = (TextView) mAct.findViewById(R.id.label_audio_file); // FIXME: this should likely go to trash -- but, is now useful for temporary testing 
		
        // Reset waveform variables, create widget, and set listener.
        mAudioWaveformKeyDown = false;
        mAudioWaveformSoundFile = null;
        mAudioWaveformMaxPos = 0;

        mAudioWaveformView = (WaveformView) mAct.findViewById(R.id.wave_form);
        mAudioWaveformView.setListener(waveformListener);
    	Log.d(LOGTAG, "onResume() 99");		
    } // END onResume()
    
    
    
    @Override
    public void onStop() {
    	Log.d(LOGTAG, "onStop() 00 bound_flag: " + mAudioServiceBound);
    	Log.d(LOGTAG, "onStop() 05 mAudioService: " + mAudioService );		
    	super.onStop();
    	if (null != mAudioService) { // FIXME: possibly merge this if block into the block above
    		int serviceState = mAudioService.getStateBnd();
    		// FIXME: a bit of concern, that the state/termination of the audio service is controlled from here, outside... is this right? how better?
    		if ( (MPQAudioService.FLG_SRV_STATE_PLAY != serviceState) &&
    				(MPQAudioService.FLG_SRV_STATE_PAUSE_PLAY != serviceState) &&
    				(MPQAudioService.FLG_SRV_STATE_REC != serviceState) &&
    				(MPQAudioService.FLG_SRV_STATE_PAUSE_REC != serviceState) ) {
    			//  (MPQAudioService.FLG_SRV_STATE_READY != serviceState) ) {
// FIXME: HACK here -- to kill the app if in READY state only ... -- should probably save this in a savedInstance Bundle... 
    			Intent intent = new Intent(mAct, MPQAudioService.class);
            	mAct.stopService(intent); // FIXME: carefully review if this belongs here ... currently it's stopping running services -- would it stop a foreground service???  ... maybe check the state of the service before stopping it...  however -- this blurs where logic/control for the service resides ... inside, or here outside?    			
    		}
    	}
    	if (mAudioServiceBound) {
    		mAudioService.delListenerBnd(getTag());
    		mAct.unbindService(mAServiceConnection);
    		mAudioServiceBound = false;
    	}
    } // END onStop()
    
    

	/*****************************************
	 * Private helper bind-connection callbacks and other Listeners.
	 */
	
    /**
     * Callbacks for service binding, passed to bindService().
     */
    private ServiceConnection mAServiceConnection = new ServiceConnection() {
    	public void onServiceConnected(ComponentName className, IBinder service_binder) {
    		MPQAudioService.LocalBinder binder = (MPQAudioService.LocalBinder) service_binder;
    		mAudioService = binder.getService();
    		mAudioService.setListenerBnd(aserviceListener, getTag());
    		mAudioServiceBound = true;
    		
    		mAudioServiceState = mAudioService.getStateBnd();
    		updateWidgetsLook();
			Log.d(LOGTAG, "mAServiceConnection onServiceConnected() 99 got State: " + mAudioService.getStateBnd());
    	} // END onServiceConnected()
    	public void onServiceDisconnected(ComponentName className) {
    		mAudioServiceBound = false;
			Log.d(LOGTAG, "mAServiceConnection onServiceDisconnected() 99 classname: " + className);
    	}
    }; // END ServiceConnection mAServiceConnection
    
    
    
	/**
	 * MPQ Audio Service Listener.
	 */
	private MPQAudioService.MPQAudioServiceListener aserviceListener = new MPQAudioService.MPQAudioServiceListener() {
		public void onMPQAudioServiceStateChange(int newState) {
			Log.d(LOGTAG, "aserviceListener  onMPQAudioServiceStateChange() 00  new State: " + newState);
			// Trigger UI updates.	
			mAudioServiceState = newState;
			updateWidgetsLook();
		} // END onMPQAudioServiceStateChange();
		
		public void onMPQAudioServiceAudioChange(long newPos, long fileLen, int newVol) {
			Log.d(LOGTAG, "aserviceListener  onMPQAudioServiceAudioChange() 00  new Pos: " + newPos + " new Vol: " + newVol);
			// for now - Update cursor position text , and or total length.
			// later - update mic volume - possibly audio wave position or pic
			boolean lupdateFlag = false;
			if ( (newPos >= 0) && (newPos != mAudioServicePos) ) {
				mAudioServicePos = newPos;
				lupdateFlag = true;				
			}
			if ( (fileLen >= 0) && (fileLen != mAudioServiceLen) ) {
				mAudioServiceLen = fileLen;
				lupdateFlag = true;
			}
			if ( (newVol >= 0) && (newVol != mAudioServiceVol) ) {
				mAudioServiceVol = newVol;
				lupdateFlag = true;
			}

			if (true == lupdateFlag) 
				updateAudioWidgetsLook();
		} // END onMPQAudioServiceAudioChange()
		
		public void onMPQAudioServiceAADone(String filename, long audioLenMS, long noiseLenMS) {
			Log.d(LOGTAG, "aserviceListener  onMPQAudioServiceAADone() 00  filename: " + filename + " audio: " + audioLenMS + " noise: " + noiseLenMS);
			// No other action for now here.
		} // END onMPQAudioServiceAADone()
	}; // END MPQAudioService.MPQAudioServiceListener aserviceListener

    

	/*****************************************
	 * Private button Listeners, for transport controls: TC/Play/Pause/Rec, Stop, Practice/Rec, Volume seek bar.
	 */
	
	/**
	 * TC session button listener.
	 * Button used for transport control: record, play, pause toggle buttons.
	 */
	private OnClickListener clickerTC = new OnClickListener() {
		public void onClick(View v) {
			Log.d(LOGTAG, "TC onClick() 00  state: " + mAudioServiceState);
			Intent intent = new Intent(v.getContext(), MPQAudioService.class);
			if ( (MPQAudioService.FLG_SRV_STATE_REC == mAudioServiceState) ||
					(MPQAudioService.FLG_SRV_STATE_PAUSE_REC == mAudioServiceState) ) {
				intent.setAction(MPQAudioService.ACTION_PAUSE_REC);
			} else if (MPQAudioService.FLG_SRV_STATE_READY == mAudioServiceState) {
				intent.setAction(MPQAudioService.ACTION_PLAY);
			} else if ( (MPQAudioService.FLG_SRV_STATE_PLAY == mAudioServiceState) ||
					(MPQAudioService.FLG_SRV_STATE_PAUSE_PLAY == mAudioServiceState) ) {
				intent.setAction(MPQAudioService.ACTION_PAUSE_PLAY);
			}
			mAct.startService(intent);
			Log.d(LOGTAG, "TC onClick() 99");
		}
	}; // END OnClickListener clickerTC
	
	/**
	 * Stop session button listener.
	 */
	private OnClickListener clickerStop = new OnClickListener() {
		public void onClick(View v) {
			Log.d(LOGTAG, "STOP onClick() 00  state: " + mAudioServiceState);
			Intent intent = new Intent(v.getContext(), MPQAudioService.class);
			intent.setAction(MPQAudioService.ACTION_STOP);
			mAct.startService(intent);
			Log.d(LOGTAG, "STOP onClick() 99");
		}
	}; // END OnClickListener clickerStop
	
	/**
	 * Practice session button listener.
	 * If currently recording (or paused from recording) then this button will stop; 
	 * otherwise it starts a new recording.
	 */
	private OnClickListener clickerPractice = new OnClickListener() {
		public void onClick(View v) {
			Log.d(LOGTAG, "PRACTICE/REC onClick() 00  state: " + mAudioServiceState);
			Intent intent = new Intent(v.getContext(), MPQAudioService.class);
			// If currently recording (or paused from recording) then this button will stop; otherwise it starts a new recording.
			if ( (MPQAudioService.FLG_SRV_STATE_REC == mAudioServiceState) ||
					(MPQAudioService.FLG_SRV_STATE_PAUSE_REC == mAudioServiceState) ){
				intent.setAction(MPQAudioService.ACTION_STOP);
			} else {
				intent.setAction(MPQAudioService.ACTION_REC);

				// Pass filename via "extra-bundle"  
	        	String lfilename = "/mpq/testAudioService01.wav";
	        	mLabelFilename.setText("filename: " + lfilename);
	        	Bundle lextraArgs = new Bundle();
	        	lextraArgs.putString(MPQAudioService.ARG_FILENAME, lfilename);
	        	intent.putExtras(lextraArgs);
			}
			mAct.startService(intent);
			Log.d(LOGTAG, "PRACTICE/REC onClick() 99");
		}
	}; // END OnClickListener clickerPractice

	
	/**
	 * Volume control SeekBar listener.
	 * NOTE: *** NOT CONNECTED TO LISTENER ABOVE ***
	 */
//	private SeekBar.OnSeekBarChangeListener seekerVolControl = new SeekBar.OnSeekBarChangeListener() {
//		public void onProgressChanged(SeekBar seekBar, int progress, boolean fromUser) {
//			Log.d(LOGTAG, "seekerVolC onProgressChanged() 00  progress: " + progress);
//			// stop the transport... stop foreground service...
//			mLabelVolValue.setText(String.valueOf(progress));
//			Log.d(LOGTAG, "seekerVolC onProgressChanged() 99");
//		}	
//		public void onStartTrackingTouch(SeekBar seekBar) {
//			Log.d(LOGTAG, "seekerVolC onStartTrackingTouch() 00  ");
//			// nothing for now		
//		}
//		public void onStopTrackingTouch(SeekBar seekBar) {
//			Log.d(LOGTAG, "seekerVolC onStopTrackingTouch() 00  ");
//			// nothing for now
//		}
//	}; // END SeekBar.OnSeekBarChangeListener seekerVolControl
	
	
	/*****************************************
	 * Private helper methods to update UI: look of buttons, audio widgets.
	 */
	
    /**
     * Update various Audio UI widgets text and state.
     * {@link #mAudioServiceState} is checked for the current state.
     * {@link #mAudioServiceLen} is updated if a new length is found.
     * 
     * UI widgets updated includes: Start/Stop Practice Session, TC toggle, 
     * Stop, and Volume control labels.
     * This aligns with the MPQAudioService design and state diagram.
     */
	private void updateWidgetsLook() {
		Log.d(LOGTAG, "updateWidgetsLook() 00  current mState: " + mAudioServiceState);
		
		// Exit without updating UI, if the fragment is not attached to an activity.
		Activity la = getActivity();
		if (null == la) {
			Log.e(LOGTAG, "updateWidgetsLook() 05->99  EXIT , no current activity");
			return;
		}

		// Show, hide, and update text for various UI buttons.
		if (MPQAudioService.FLG_SRV_STATE_INIT == mAudioServiceState) {
			mTCButton.setText(R.string.button_aui_play);
			mTCButton.setEnabled(false);
			mStopButton.setEnabled(false);
			mLabelAudioPos.setVisibility(View.INVISIBLE);
			
			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mLabelVolControl.setText(getString(R.string.label_aui_vol_control));
//			mPracticeButton.setText(getString(R.string.button_practice_ready)); 
//			mPracticeButton.setEnabled(true);
		} else if (MPQAudioService.FLG_SRV_STATE_READY == mAudioServiceState)  {
			mTCButton.setText(R.string.button_aui_play);
			mTCButton.setEnabled(true);
			mStopButton.setEnabled(false);
			mLabelAudioPos.setVisibility(View.VISIBLE);
			
			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mLabelVolControl.setText(getString(R.string.label_aui_vol_control));
//			mPracticeButton.setText(getString(R.string.button_practice_ready)); 
//			mPracticeButton.setEnabled(true);
		} else if (MPQAudioService.FLG_SRV_STATE_REC == mAudioServiceState) {
			// Recording
			mTCButton.setText(R.string.button_aui_pause);
			mTCButton.setEnabled(true);
			mStopButton.setEnabled(true);
			mLabelAudioPos.setVisibility(View.INVISIBLE);

			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mLabelVolControl.setText(getString(R.string.label_aui_mic_control));
//			mPracticeButton.setText(getString(R.string.button_practice_inprocess));			
//			mPracticeButton.setEnabled(true);
		} else if (MPQAudioService.FLG_SRV_STATE_PAUSE_REC == mAudioServiceState) {
			mTCButton.setText(R.string.button_aui_rec);
			mTCButton.setEnabled(true);
			mStopButton.setEnabled(true);
			mLabelAudioPos.setVisibility(View.INVISIBLE);

			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mLabelVolControl.setText(getString(R.string.label_aui_mic_control));
//			mPracticeButton.setText(getString(R.string.button_practice_inprocess)); 
//			mPracticeButton.setEnabled(true); 
		} else if (MPQAudioService.FLG_SRV_STATE_PAUSE_PLAY == mAudioServiceState) {
			mTCButton.setText(R.string.button_aui_play);
			mTCButton.setEnabled(true);
			mStopButton.setEnabled(true);
			mLabelAudioPos.setVisibility(View.VISIBLE);

			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mLabelVolControl.setText(getString(R.string.label_aui_vol_control));
//			mPracticeButton.setText(getString(R.string.button_practice_ready)); 
//			mPracticeButton.setEnabled(false); 
		} else if (MPQAudioService.FLG_SRV_STATE_PLAY == mAudioServiceState) {
			mTCButton.setText(R.string.button_aui_pause);
			mTCButton.setEnabled(true);
			mStopButton.setEnabled(true);
			mLabelAudioPos.setVisibility(View.VISIBLE);
			
			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mLabelVolControl.setText(getString(R.string.label_aui_vol_control));
//			mPracticeButton.setText(getString(R.string.button_practice_ready)); 
//			mPracticeButton.setEnabled(false); 
		} else if (MPQAudioService.FLG_SRV_STATE_STOP == mAudioServiceState) {
			// Leave previous mTCButton text.
			mTCButton.setEnabled(false);
			mStopButton.setEnabled(false);
			mLabelAudioPos.setVisibility(View.INVISIBLE);
			
			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mLabelVolControl.setText(getString(R.string.label_aui_vol_control));
//			mPracticeButton.setText(getString(R.string.button_practice_ready)); 
//			mPracticeButton.setEnabled(false); 
		} 
		  else {
			Log.e(LOGTAG, "updateWidgetsLook() 80  (ignoring) UNHANDLED State: " + mAudioServiceState);
		}
		
		// For all states (FIXME: confirm that's OK) check the current file, and if different, update the UI text and waveform.
    	String lFileName = "";
    	if (null != mAudioService) { 
        	lFileName = mAudioService.getFilenameBnd();
    		Log.d(LOGTAG, "updateWidgetsLook() 82  current filename: " + mAudioServiceName + " new filename: " + lFileName);
        	if ( (lFileName != null) && (lFileName != mAudioServiceName) ) {
        		mAudioServiceName = lFileName;
            	mLabelFilename.setText("filename: " + lFileName);
        		
        		// Update mAudioWaveformView widget. // FIXME: Bug with displaying new recording, skip for now. 
                mAudioWaveformMaxPos = 0;
                if (MPQAudioService.FLG_SRV_STATE_REC != mAudioServiceState) {
                	waveformLoadFromFile();
                }
                
// FIXME: this may belong in the audioWidgetUpdate, it's likely not getting called here...
        		Log.d(LOGTAG, "updateWidgetsLook() 84  SoundFile: " + mAudioWaveformSoundFile);
                if (mAudioWaveformSoundFile != null) {
                    DisplayMetrics metrics = new DisplayMetrics();
                    mAct.getWindowManager().getDefaultDisplay().getMetrics(metrics);
                    mAudioWaveformDensity = metrics.density;
                    
                    mAudioWaveformView.setSoundFile(mAudioWaveformSoundFile);
                    mAudioWaveformView.recomputeHeights(mAudioWaveformDensity);
                    mAudioWaveformMaxPos = mAudioWaveformView.maxPos();
                }	
        	}
// FIXME: testing this out here also... review and adjust        	
Log.d(LOGTAG, "updateWidgetsLook() 84BBBBB SFile: " + mAudioWaveformSoundFile);
//if (mAudioWaveformSoundFile != null) {
//    DisplayMetrics metrics = new DisplayMetrics();
//    mAct.getWindowManager().getDefaultDisplay().getMetrics(metrics);
//    mAudioWaveformDensity = metrics.density;
//    
//    mAudioWaveformView.setSoundFile(mAudioWaveformSoundFile);
//    mAudioWaveformView.recomputeHeights(mAudioWaveformDensity);
//    mAudioWaveformMaxPos = mAudioWaveformView.maxPos();
//}	        	
        	        	
        	// Update file length and cursor positions (this is called infrequently).
        	long lFileLenPos = mAudioService.getFileLengthMSBnd();
    		Log.d(LOGTAG, "updateWidgetsLook() 86  current len: " + mAudioServiceLen + " new fileLen: " + lFileLenPos);
        	if ( (lFileLenPos >= 0) && (lFileLenPos != mAudioServiceLen) ) {
        		mAudioServiceLen = lFileLenPos;
        	}        	
        	lFileLenPos = mAudioService.getFileposBnd();
        	Log.d(LOGTAG, "updateWidgetsLook() 87  current pos: " + mAudioServicePos + " new pos: " + lFileLenPos);
        	if ( (lFileLenPos >= 0) && (lFileLenPos != mAudioServicePos) ) {
        		mAudioServicePos = lFileLenPos;
        	}        	
    	}
    	
    	// Follow up with update of Audio related UI elements.
		updateAudioWidgetsLook();
		
	} // END updateWidgetsLook()
	
	
    /**
     * Update frequently changing Audio UI widgets, such as cursor position and volume.
     * {@link #mAudioServiceState} is checked for the current state.
     * {@link #mAudioServicePos} is checked for the current position.
     * {@link #mAudioServiceVol} is checked for the current volume.
     * 
     * UI widgets updated includes: cursor position, file length, volume level.
     * This aligns with the MPQAudioService design and state diagram.
     */
	private void updateAudioWidgetsLook() {
		Log.d(LOGTAG, "updateAudioWidgetsLook() 00  current mState: " + mAudioServiceState + " pos: " + mAudioServicePos + " len: " + mAudioServiceLen + " vol: " + mAudioServiceVol);

		if (MPQAudioService.FLG_SRV_STATE_READY == mAudioServiceState) {
			mLabelAudioPos.setText( SessionsDB.longSessionMSecString(mAudioServicePos) + "/" );
		} else if ( (MPQAudioService.FLG_SRV_STATE_REC == mAudioServiceState) || 
				(MPQAudioService.FLG_SRV_STATE_PAUSE_REC == mAudioServiceState) ) {
			// No mLabelAudioPos while recording.
//// FIXME: confirm that this is not needed -- probably toss logging below also --			mAudioServiceLen = mAudioServicePos;			
			Log.d(LOGTAG, "updateAudioWidgetsLook() 20  current mAudioServiceLen: " +mAudioServiceLen + " Pos: " + mAudioServicePos);
		} else if ( (MPQAudioService.FLG_SRV_STATE_PLAY == mAudioServiceState) ||
				(MPQAudioService.FLG_SRV_STATE_PAUSE_PLAY == mAudioServiceState) ) {
			mLabelAudioPos.setText( SessionsDB.longSessionMSecString(mAudioServicePos) + "/");
		}
		
//		Log.d(LOGTAG, "updateAudioWidgetsLook() 84  SoundFile: " + mAudioWaveformSoundFile);
//        if (mAudioWaveformSoundFile != null) {
//            DisplayMetrics metrics = new DisplayMetrics();
//            getWindowManager().getDefaultDisplay().getMetrics(metrics);
//            mAudioWaveformDensity = metrics.density;
//            
//            mAudioWaveformView.setSoundFile(mAudioWaveformSoundFile);
//            mAudioWaveformView.recomputeHeights(mAudioWaveformDensity);
//            mAudioWaveformMaxPos = mAudioWaveformView.maxPos();
//        }

		
		mLabelAudioLen.setText( SessionsDB.longSessionMSecString(mAudioServiceLen) );
//		mLabelVolValue.setText(String.valueOf(mVolControl.getProgress()));
	} // END updateAudioWidgetsLook()
	
	
	
	
	
	
	/*****************************************
	 * Waveform View Listener, followed by helper methods to update UI and load files. -- credit: Ringdroid
	 */
		
	/**
	 * Waveform UI widget listener.
	 */
	private WaveformView.WaveformListener waveformListener = new WaveformView.WaveformListener() {	
		/**
		 * Every time we get a message that our waveform drew, see if we need to
		 * animate and trigger another redraw.
		 */
		public void waveformDraw() {
	    	Log.d(LOGTAG, "waveformListener waveformDraw() 00 = serviceState: " + mAudioServiceState);
			mAudioWaveformWidth = mAudioWaveformView.getMeasuredWidth();
			if (mAudioWaveformOffsetGoal != mAudioWaveformOffset && !mAudioWaveformKeyDown)
				waveformUpdateDisplay();
			else if (MPQAudioService.FLG_SRV_STATE_PLAY == mAudioServiceState) {
				waveformUpdateDisplay();
			}
	//		else if (mFlingVelocity != 0) {
	//			waveformUpdateDisplay();
	//		}
		} // END waveformDraw()
	
	//	public void waveformZoomIn() {
	//		mAudioWaveformView.zoomIn();
	//		mStartPos = mAudioWaveformView.getStart();
	//		mEndPos = mAudioWaveformView.getEnd();
	//		mMaxPos = mAudioWaveformView.maxPos();
	//		mOffset = mAudioWaveformView.getOffset();
	//		mOffsetGoal = mOffset;
	//////		enableZoomButtons();
	//		waveformUpdateDisplay();
	//	}
	//
	//	public void waveformZoomOut() {
	//		mAudioWaveformView.zoomOut();
	//		mStartPos = mAudioWaveformView.getStart();
	//		mEndPos = mAudioWaveformView.getEnd();
	//		mMaxPos = mAudioWaveformView.maxPos();
	//		mOffset = mAudioWaveformView.getOffset();
	//		mOffsetGoal = mOffset;
	//////		enableZoomButtons();
	//		waveformUpdateDisplay();
	//	}
	
		public void waveformTouchStart(float x) {
	    	Log.d(LOGTAG, "waveformListener waveformTouchStart() 00 = x: " + x);
	//		mTouchDragging = true;
	//		mTouchStart = x;
	//		mTouchInitialOffset = mOffset;
	//		mFlingVelocity = 0;
	//		mWaveformTouchStartMsec = System.currentTimeMillis();
		} // END waveformTouchStart()
	
	//	public void waveformTouchMove(float x) {
	//		mOffset = trap((int)(mTouchInitialOffset + (mTouchStart - x)));
	//		updateDisplay();
	//	}
	//
	//	public void waveformTouchEnd() {
	//		mTouchDragging = false;
	//		mOffsetGoal = mOffset;
	//
	//		long elapsedMsec = System.currentTimeMillis() -
	//				mWaveformTouchStartMsec;
	//		if (elapsedMsec < 300) {
	//			if (mIsPlaying) {
	//				int seekMsec = mWaveformView.pixelsToMillisecs(
	//						(int)(mTouchStart + mOffset));
	//				if (seekMsec >= mPlayStartMsec &&
	//						seekMsec < mPlayEndMsec) {
	//					mPlayer.seekTo(seekMsec - mPlayStartOffset);
	//				} else {
	//					handlePause();
	//				}
	//			} else {
	//				onPlay((int)(mTouchStart + mOffset));
	//			}
	//		}
	//	}
	//
	//	public void waveformFling(float vx) {
	//		mTouchDragging = false;
	//		mOffsetGoal = mOffset;
	//		mFlingVelocity = (int)(-vx);
	//		updateDisplay();
	//	}
	}; // END WaveformView.WaveformListener waveformListener
		
	
	/**
	 * WTF does this do??? FIXME:
	 */
    private void waveformUpdateDisplay() {
    	Log.d(LOGTAG, "waveformUpdateDisplay() 00 = serviceState: " + mAudioServiceState);

		if (MPQAudioService.FLG_SRV_STATE_PLAY == mAudioServiceState) {
            long now = mAudioServicePos;
            int frames = mAudioWaveformView.millisecsToPixels( (int) now); // FIXME: hack cast over here -- clean this up
            mAudioWaveformView.setPlayback(frames);
 ////////           setOffsetGoalNoUpdate(frames - mAudioWaveformWidth / 2);
//            if (now >= mPlayEndMsec) {
//                handlePause();
//            }
        }

//        if (!mTouchDragging) {
//            int offsetDelta;
//
//            if (mFlingVelocity != 0) {
//                float saveVel = mFlingVelocity;
//
//                offsetDelta = mFlingVelocity / 30;
//                if (mFlingVelocity > 80) {
//                    mFlingVelocity -= 80;
//                } else if (mFlingVelocity < -80) {
//                    mFlingVelocity += 80;
//                } else {
//                    mFlingVelocity = 0;
//                }
//
//                mOffset += offsetDelta;
//
//                if (mOffset + mWidth / 2 > mMaxPos) {
//                    mOffset = mMaxPos - mWidth / 2;
//                    mFlingVelocity = 0;
//                }
//                if (mOffset < 0) {
//                    mOffset = 0;
//                    mFlingVelocity = 0;
//                }
//                mOffsetGoal = mOffset;
//            } else {
//                offsetDelta = mOffsetGoal - mOffset;
//
//                if (offsetDelta > 10)
//                    offsetDelta = offsetDelta / 10;
//                else if (offsetDelta > 0)
//                    offsetDelta = 1;
//                else if (offsetDelta < -10)
//                    offsetDelta = offsetDelta / 10;
//                else if (offsetDelta < 0)
//                    offsetDelta = -1;
//                else
//                    offsetDelta = 0;
//
//                mOffset += offsetDelta;
//            }
//        } // END if (!mTouchDragging) 

    	Log.d(LOGTAG, "waveformUpdateDisplay() 50 = StartPos: " + mAudioWaveformStartPos + " EndPos: " + mAudioWaveformEndPos + " Offset: " + mAudioWaveformOffset);
        mAudioWaveformView.setParameters(mAudioWaveformStartPos, mAudioWaveformEndPos, mAudioWaveformOffset);
        mAudioWaveformView.invalidate();

//        mStartMarker.setContentDescription(
//            getResources().getText(R.string.start_marker) + " " +
//            formatTime(mStartPos));
//        mEndMarker.setContentDescription(
//            getResources().getText(R.string.end_marker) + " " +
//            formatTime(mEndPos));

//        int startX = mStartPos - mOffset - mMarkerLeftInset;
//        if (startX + mStartMarker.getWidth() >= 0) {
//        	if (!mStartVisible) {
//        		// Delay this to avoid flicker
//        		mHandler.postDelayed(new Runnable() {
//        			public void run() {
//        				mStartVisible = true;
//        				mStartMarker.setAlpha(255);
//        			}
//        		}, 0);
//        	}
//        } else {
//        	if (mStartVisible) {
//        		mStartMarker.setAlpha(0);
//        		mStartVisible = false;
//        	}
//        	startX = 0;
//        }

//        int endX = mEndPos - mOffset - mEndMarker.getWidth() +
//        		mMarkerRightInset;
//        if (endX + mEndMarker.getWidth() >= 0) {
//        	if (!mEndVisible) {
//        		// Delay this to avoid flicker
//        		mHandler.postDelayed(new Runnable() {
//        			public void run() {
//        				mEndVisible = true;
//        				mEndMarker.setAlpha(255);
//        			}
//        		}, 0);
//        	}
//        } else {
//        	if (mEndVisible) {
//        		mEndMarker.setAlpha(0);
//        		mEndVisible = false;
//        	}
//        	endX = 0;
//        }

//        mStartMarker.setLayoutParams(
//            new AbsoluteLayout.LayoutParams(
//                AbsoluteLayout.LayoutParams.WRAP_CONTENT,
//                AbsoluteLayout.LayoutParams.WRAP_CONTENT,
//                startX,
//                mMarkerTopOffset));
//
//        mEndMarker.setLayoutParams(
//            new AbsoluteLayout.LayoutParams(
//                AbsoluteLayout.LayoutParams.WRAP_CONTENT,
//                AbsoluteLayout.LayoutParams.WRAP_CONTENT,
//                endX,
//                mWaveformView.getMeasuredHeight() -
//                mEndMarker.getHeight() - mMarkerBottomOffset));
    } // END waveformUpdateDisplay()

	  

     private void waveformLoadFromFile() {
 		Log.d(LOGTAG, "waveformLoadFromFile() 00  ");

  //////       mFile = new File(mAudioServiceName);
  //       mExtension = getExtensionFromFilename(mAudioServiceName);

//         SongMetadataReader metadataReader = new SongMetadataReader(
//             this, mFilename);
//         mTitle = metadataReader.mTitle;
//         mArtist = metadataReader.mArtist;
//         mAlbum = metadataReader.mAlbum;
//         mYear = metadataReader.mYear;
//         mGenre = metadataReader.mGenre;
 //
//         String titleLabel = mTitle;
//         if (mArtist != null && mArtist.length() > 0) {
//             titleLabel += " - " + mArtist;
//         }
//         setTitle(titleLabel);

 ////// FIXME: review and delete (along with declaration) --        mAudioWaveformLoadingStartTime = System.currentTimeMillis();
         mAudioWaveformLoadingLastUpdateTime = System.currentTimeMillis();
         mAudioWaveformLoadingKeepGoing = true;
         mAudioWaveformProgressDialog = new ProgressDialog(mAct );
         mAudioWaveformProgressDialog.setProgressStyle(ProgressDialog.STYLE_HORIZONTAL);
         mAudioWaveformProgressDialog.setTitle(R.string.progress_dialog_loading);
         mAudioWaveformProgressDialog.setCancelable(true);
         mAudioWaveformProgressDialog.setOnCancelListener(
             new DialogInterface.OnCancelListener() {
                 public void onCancel(DialogInterface dialog) {
                 	mAudioWaveformLoadingKeepGoing = false;
                 }
             });
         mAudioWaveformProgressDialog.show();

         final CheapSoundFile.ProgressListener listener =
             new CheapSoundFile.ProgressListener() {
                 public boolean reportProgress(double fractionComplete) {
                     long now = System.currentTimeMillis();
                     if (now - mAudioWaveformLoadingLastUpdateTime > 100) {
                         mAudioWaveformProgressDialog.setProgress(
                             (int)(mAudioWaveformProgressDialog.getMax() *
                                   fractionComplete));
                         mAudioWaveformLoadingLastUpdateTime = now;
                     }
                     return mAudioWaveformLoadingKeepGoing;
                 }
             };

//         // Create the MediaPlayer in a background thread
//         new Thread() {
//             public void run() {
//                 mCanSeekAccurately = SeekTest.CanSeekAccurately(
//                     getPreferences(Context.MODE_PRIVATE));
 //
//                 System.out.println("Seek test done, creating media player.");
//                 try {
//                     MediaPlayer player = new MediaPlayer();
//                     player.setDataSource(mFile.getAbsolutePath());
//                     player.setAudioStreamType(AudioManager.STREAM_MUSIC);
//                     player.prepare();
//                     mPlayer = player;
//                 } catch (final java.io.IOException e) {
//                     Runnable runnable = new Runnable() {
//                         public void run() {
//                             handleFatalError(
//                                 "ReadError",
//                                 getResources().getText(R.string.read_error),
//                                 e);
//                         }
//                     };
//                     mHandler.post(runnable);
//                 };
//             }
//         }.start();

         // Load the sound file in a background thread
 //FIXME: review the member variable access by these threads, it does not feel safe... 
         new Thread() { 
             public void run() { 
// FIXME: check for cache (or create cache if missing)            	 
                 Log.d(LOGTAG, "waveformLoadFromFile() thread run() 00 to load cheap soundfile");
                 try {
//////                 	File lFile = new File(Environment.getExternalStorageDirectory().getAbsolutePath() + mAudioServiceName);
                	 String lFile = Environment.getExternalStorageDirectory().getAbsolutePath() + mAudioServiceName;
                	 mAudioWaveformSoundFile = CheapSoundFile.create(lFile, listener);
                 } catch (final Exception e) {
                     Log.e(LOGTAG, "waveformLoadFromFile() thread run() Exception: " + e.getLocalizedMessage() );
                     e.printStackTrace();
  //                   mInfo.setText(e.toString());
                     // FIXME: handle this exception -- do we throw the exception again??? or do we kill/finish ourselves???
                 }
                 Log.d(LOGTAG, "waveformLoadFromFile() thread run() 60 got sound file, about to dismiss and call finish()");
                 mAudioWaveformProgressDialog.dismiss(); 
                 if (mAudioWaveformLoadingKeepGoing) {
                     Runnable runnable = new Runnable() {
                             public void run() {
                                 Log.d(LOGTAG, "waveformLoadFromFile() thread run() 88 SUB thread to call finishOpen()");
                                 waveformFinishLoadFromFile();
                             }
                         };
                     mFragHandler.post(runnable);
                         
                 } 
             } 
         }.start();
 		Log.d(LOGTAG, "waveformLoadFromFile() 99  ");
     } // END waveformLoadFromFile()

     private void waveformFinishLoadFromFile() {
  		Log.d(LOGTAG, "waveformFinishLoadFromFile() 00  file: " + mAudioServiceName);
  		
  		// Exit early if there is not a valid sound file.
  		if (null == mAudioWaveformSoundFile) {
  			// FIXME: display the MPQ logo or some other picture instead.  Disable/hide audio controls.
  			return;
  		}
  		
        mAudioWaveformView.setSoundFile(mAudioWaveformSoundFile);
        mAudioWaveformView.recomputeHeights(mAudioWaveformDensity);

        mAudioWaveformMaxPos = mAudioWaveformView.maxPos();
//         mAudioWaveformLastDisplayedStartPos = -1;
//         mAudioWaveformLastDisplayedEndPos = -1;

  //       mTouchDragging = false;

        mAudioWaveformOffset = 0;
        mAudioWaveformOffsetGoal = 0;
 ///////        mFlingVelocity = 0;
        waveformResetPositions();
        if (mAudioWaveformEndPos > mAudioWaveformMaxPos) {
         	mAudioWaveformEndPos = mAudioWaveformMaxPos;
        }


 // FIXME: review and delete this... does the below work enough???        updateDisplay();
  
        Log.d(LOGTAG, "waveformFinishLoadFromFile() 97");
        updateWidgetsLook();
   		Log.d(LOGTAG, "waveformFinishLoadFromFile() 99");
     } // END waveformFinishLoadFromFile()
 	
     private void waveformResetPositions() {
     	mAudioWaveformStartPos = mAudioWaveformView.secondsToPixels(0.0);
     	mAudioWaveformEndPos = mAudioWaveformView.secondsToPixels(15.0);
     	//mAudioWaveformEndPos = mAudioWaveformView.getEnd();
     }
    
} // END class MPQAudioFragment
